#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* Author: Maurice Castro */
/* WWW: http://www.castro.aus.net/~maurice */
/* E-mail: http://www.castro.aus.net/~maurice/mail */
/* $Id: triemaker.c,v 1.8 2010/03/10 03:28:04 maurice Exp $ */

/* this program has a format similar to yacc and lex: files are */
/* divided into 3 parts. The first part contains directives and code */
/* to be output before the part computed by the program. The 2nd part */
/* contains the input material arranged as a two column list which is */
/* output two tables the first table is a base 10 trie and the second */
/* table is the set of strings that are terminal nodes of the trie. */
/* The final part is code that is output as entered after the computed */
/* tables. The parts are separated by lines beginning with '%%' */

#define SEP "%%"
#define SEPLEN 2
#define LITBEGIN "%{"
#define LITEND "}%"
#define LITLEN 2

/* formating strings */
char *comment_open = "/* ";
char *comment_close = " */";
char *column_open = "{\n";
char *column_close = "\n}";
char *column_sep = ",\n";
char *row_open = "{";
char *row_close = "}";
char *row_sep = ",";
char *trie_tab_decl = "int trie_tab[%d][11] =\n";
char *out_tab_decl = "char *out_tab[%d] =\n";
char *term_tab_decl = "struct terms { char *term; int out; }; struct terms term_tab[%d] =\n";
char *statement_sep = ";";
char *string_delim = "\"";
char *out_err = "NULL";

char *c_mode_comment_open = "/* ";
char *c_mode_comment_close = " */";
char *c_mode_column_open = "{\n";
char *c_mode_column_close = "\n}";
char *c_mode_column_sep = ",\n";
char *c_mode_row_open = "{";
char *c_mode_row_close = "}";
char *c_mode_row_sep = ",";
char *c_mode_trie_tab_decl = "int trie_tab[%d][11] =\n";
char *c_mode_out_tab_decl = "char *out_tab[%d] =\n";
char *c_mode_term_tab_decl = "struct terms { char *term; int out; }; struct terms term_tab[%d] =\n";
char *c_mode_statement_sep = ";";
char *c_mode_out_err = "NULL";
char *c_mode_string_delim = "\"";

char *perl_mode_comment_open = "# ";
char *perl_mode_comment_close = "";
char *perl_mode_column_open = "(\n";
char *perl_mode_column_close = ")\n";
char *perl_mode_column_sep = ",\n";
char *perl_mode_row_open = "[";
char *perl_mode_row_close = "]";
char *perl_mode_row_sep = ",";
char *perl_mode_trie_tab_decl = "@trie_tab =\n";
char *perl_mode_out_tab_decl = "@out_tab =\n";
char *perl_mode_term_tab_decl = "@term_tab =\n";
char *perl_mode_statement_sep = ";";
char *perl_mode_out_err  = "";
char *perl_mode_string_delim = "\"";

char *php_mode_comment_open = "#";
char *php_mode_comment_close = "";
char *php_mode_column_open = "array(\n";
char *php_mode_column_close = ")\n";
char *php_mode_column_sep = ",\n";
char *php_mode_row_open = "array(";
char *php_mode_row_close = ")";
char *php_mode_row_sep = ",";
char *php_mode_trie_tab_decl = "$trie_tab =\n";
char *php_mode_out_tab_decl = "$out_tab =\n";
char *php_mode_term_tab_decl = "$term_tab =\n";
char *php_mode_statement_sep = ";";
char *php_mode_out_err = "";
char *php_mode_string_delim = "\"";

/* debugging data */
extern int errno;
char *prognam = "triemaker";
int lineno = 0;
int debug_mode = 0;

/* input and output files */
FILE *fout;
FILE *fin;
int bufsiz = 255;

int no_duplicates = 0;

/* string table and in memory trie structures */
typedef struct stringTab *stringTab_t;

struct stringTab
{
	int n;
	char *s;
	stringTab_t next;
};

enum TRIE_TYPE {TERMINAL, INTERNAL};

typedef struct inMemTrie *inMemTrie_t;

struct terminal
{
	int nt;
	char *num;
	stringTab_t value;
};

struct internal
{
	int ni;
	inMemTrie_t digit[11];
};

struct inMemTrie
{
	int n;
	enum TRIE_TYPE t;
	union
	{
		struct terminal term;
		struct internal intern;
	};
};

struct trieTab
{
	int digit[11];
};

struct termTab
{
	char *term;
	int outidx;
};

inMemTrie_t top = NULL;
stringTab_t strtab = NULL;

/* the value of zero is reserved for the n_intern and n_term values */
int nicnt = 1;
int ntcnt = 1;

/* functions */
void memchk(void *p)
{
	if (!p)
	{
		fprintf(stderr, "%s: Out of Memory at line %d\n", prognam, lineno);
		exit(1);
	}
}

stringTab_t newstringTab(stringTab_t *tab, char *s)
{
	static int cnt = 1;
	stringTab_t r;
	if (!*tab)
	{
		*tab = malloc(sizeof(*r));
		memchk(*tab);
		(*tab)->n = cnt++;
		(*tab)->s = strdup(s);
		(*tab)->next = NULL;
		memchk((*tab)->s);
		return *tab;
	}
	r = *tab;
	while (r)
	{
		if (!strcmp(s, r->s))
			return r;
		if (!r->next)
		{
			r->next = malloc(sizeof(*r));
			memchk(r->next);
			r = r->next;
			r->n = cnt++;
			r->s = strdup(s);
			r->next = NULL;
			memchk(r->s);
			return r;
		}
		r = r->next;
	}
	return NULL; /* should never reach here */
}

void dump_strtab(stringTab_t st)
{
	stringTab_t p;
	int n = 1;
	p = st;
	while(p)
	{
		n++;
		p = p->next;
	}
	fprintf(fout,out_tab_decl, n);
	fprintf(fout,column_open);
	fprintf(fout,out_err);
	p = st;
	while(p)
	{
		fprintf(fout,column_sep);
		fprintf(fout,"%s",p->s);
		p = p->next;
	}
	fprintf(fout,column_close);
	fprintf(fout,statement_sep);
	fprintf(fout,"\n");
}

inMemTrie_t insertTrie(inMemTrie_t t, int lvl, char *key, char *value)
{
	static int cnt = 0;
	inMemTrie_t r;
	int d;
	/* if the tree is empty - make a new terminal node and return it */
	if (!t)
	{
		r = calloc(1, sizeof(*r));
		memchk(r);
		r->n = cnt++;
		r->t = TERMINAL;
		r->term.nt = ntcnt++;
		r->term.num = strdup(key);
		memchk(r->term.num);
		r->term.value = newstringTab(&strtab, value);
		return r;
	}
	/* are we at a terminal node */
	if (t->t == TERMINAL)
	{
		/* you cannot add to a terminal node so we need to create an
		   internal node above it */
		r = calloc(1, sizeof(*r));
		memchk(r);
		r->n = cnt++;
		r->t = INTERNAL;
		r->intern.ni = nicnt++;
		/* attach existing terminal node to new internal node */
		if (lvl == (strlen(t->term.num)))
			r->intern.digit[10] = t;
		else
		{
			d = t->term.num[lvl] - '0';
			r->intern.digit[d] = t;
		}
		return insertTrie(r, lvl, key, value);
	}
	/* we must be at an internal node */
	if (lvl == (strlen(key)))
	{
		if (no_duplicates && t->intern.digit[10])
		{
			fprintf(stderr, "%s: Duplicate key at line %d\n", prognam, lineno);
			exit(1);
		}
		t->intern.digit[10] = insertTrie(NULL, lvl, key, value);
	}
	else
	{
		d = key[lvl] - '0';
		t->intern.digit[d] = insertTrie(t->intern.digit[d], lvl+1, key, value);
	}
	return t;
}

void propagateLeaves(inMemTrie_t t)
{
	int i;
	if (!t)
		return;
	if (t->t == INTERNAL)
	{
		if(t->intern.digit[10])
		{
			for(i=0; i<11;i++)
			{
				if(t->intern.digit[i])
				{
					if(t->intern.digit[i]->t == INTERNAL)
					{
						if(!t->intern.digit[i]->intern.digit[10])
							t->intern.digit[i]->intern.digit[10] = t->intern.digit[10];
							
					}
				}
			}
		}
		for(i=0; i<10;i++)
		{
			propagateLeaves(t->intern.digit[i]);
		}
	}
}

void dump_trie(inMemTrie_t t)
{
	int i;
	if (!t)
		return;
	if (t->t == TERMINAL)
	{
		printf(comment_open);
		printf("%d\tTERM\t%s:\t", t->n, t->term.num);
		printf("%d %s", t->term.value->n, t->term.value->s);
		printf(comment_close);
		printf("\n");
	}
	else
	{
		printf(comment_open);
		printf("%d\tINTERN \t:\t", t->n);
		for(i=0; i<11;i++)
		{
			if(t->intern.digit[i])
			{
				printf(" %d", t->intern.digit[i]->n);
			}
			else
			{
				printf (" -1");
			}
		}
		printf(comment_close);
		printf("\n");
		for(i=0; i<11;i++)
		{
			dump_trie(t->intern.digit[i]);
		}
	}
}

void trie_map(inMemTrie_t t, struct trieTab *trie_tab, struct termTab *term_tab)
{
	int i;
	int tt;
	if (!t)
		return;
	if (t->t == INTERNAL)
	{
		for(i=0; i<11;i++)
		{
			if(t->intern.digit[i])
			{
				if (t->intern.digit[i]->t == TERMINAL)
				{
					tt = t->intern.digit[i]->term.nt;
					trie_tab[t->intern.ni].digit[i] = -tt;
					term_tab[tt].term = t->intern.digit[i]->term.num;
					term_tab[tt].outidx = t->intern.digit[i]->term.value->n;
				}
				else
					trie_tab[t->intern.ni].digit[i] = t->intern.digit[i]->intern.ni;
			}
			else
			{
				trie_tab[t->intern.ni].digit[i] = 0;
			}
		}
		for(i=0; i<11;i++)
		{
			trie_map(t->intern.digit[i], trie_tab, term_tab);
		}
	}
}

char *find_trie(inMemTrie_t t, char *key)
{
	int lvl=0;
	int d;
	inMemTrie_t p;
	if (!t)
		return NULL;
	do
	{
		if (t->t == TERMINAL)
		{
			return t->term.value->s;
		}
		else
		{
			if (key[lvl])
			{
				d = key[lvl] - '0';
				p = t->intern.digit[d];
				if (p)
				{
					lvl++;
					t = p;
				}
				else
				{
					t = t->intern.digit[10];
				}
			}
			else
			{
				t = t->intern.digit[10];
			}
		}
	} while (t);
	return NULL;
}

void dump_trie_tab(struct trieTab *trie_tab, int n)
{
	int i;
	int j;
	
	fprintf(fout, trie_tab_decl, n);
	fprintf(fout, column_open);
	fprintf(fout, row_open);
	for (j=0;j <10;j++)
	{
		fprintf(fout, "%d", trie_tab[0].digit[j]);
		fprintf(fout, row_sep);
	}
	fprintf(fout, "%d", trie_tab[0].digit[10]);
	fprintf(fout, row_close);
	for (i=1; i < n; i++)
	{
		fprintf(fout, column_sep);
		fprintf(fout, row_open);
		for (j=0;j <10;j++)
		{
			fprintf(fout, "%d", trie_tab[i].digit[j]);
			fprintf(fout, row_sep);
		}
		fprintf(fout, "%d", trie_tab[i].digit[10]);
		fprintf(fout, row_close);
	}
	fprintf(fout, column_close);
	fprintf(fout, statement_sep);
	fprintf(fout, "\n");
}

void dump_term_tab(struct termTab *term_tab, int n)
{
	int i;
	
	fprintf(fout, term_tab_decl, n);
	fprintf(fout, column_open);
	fprintf(fout, row_open);
	fprintf(fout, string_delim);
	fprintf(fout, string_delim);
	fprintf(fout, row_sep);
	fprintf(fout, "0");
	fprintf(fout, row_close);
	for (i=1; i< ntcnt; i++)
	{
		fprintf(fout, column_sep);
		fprintf(fout, row_open);
		fprintf(fout, string_delim);
		if (term_tab[i].term)
			fprintf(fout, "%s", term_tab[i].term);
		fprintf(fout, string_delim);
		fprintf(fout, row_sep);
		fprintf(fout, "%d", term_tab[i].outidx);
		fprintf(fout, row_close);
	}
	fprintf(fout, column_close);
	fprintf(fout, statement_sep);
	fprintf(fout, "\n");
}

void linein(char *buf, FILE *f)
{
	fgets(buf, bufsiz, f);
	lineno++;
}

void c_mode()
{
	comment_open = c_mode_comment_open;
	comment_close = c_mode_comment_close;
	column_open = c_mode_column_open;
	column_close = c_mode_column_close;
	column_sep = c_mode_column_sep;
	row_open = c_mode_row_open;
	row_close = c_mode_row_close;
	row_sep = c_mode_row_sep;
	trie_tab_decl = c_mode_trie_tab_decl;
	out_tab_decl = c_mode_out_tab_decl;
	term_tab_decl = c_mode_term_tab_decl;
	statement_sep = c_mode_statement_sep;
	out_err = c_mode_out_err;
	string_delim = c_mode_string_delim;
}

void perl_mode()
{
	comment_open = perl_mode_comment_open;
	comment_close = perl_mode_comment_close;
	column_open = perl_mode_column_open;
	column_close = perl_mode_column_close;
	column_sep = perl_mode_column_sep;
	row_open = perl_mode_row_open;
	row_close = perl_mode_row_close;
	row_sep = perl_mode_row_sep;
	trie_tab_decl = perl_mode_trie_tab_decl;
	out_tab_decl = perl_mode_out_tab_decl;
	term_tab_decl = perl_mode_term_tab_decl;
	statement_sep = perl_mode_statement_sep;
	out_err = perl_mode_out_err;
	string_delim = perl_mode_string_delim;
}

void php_mode()
{
	comment_open = php_mode_comment_open;
	comment_close = php_mode_comment_close;
	column_open = php_mode_column_open;
	column_close = php_mode_column_close;
	column_sep = php_mode_column_sep;
	row_open = php_mode_row_open;
	row_close = php_mode_row_close;
	row_sep = php_mode_row_sep;
	trie_tab_decl = php_mode_trie_tab_decl;
	out_tab_decl = php_mode_out_tab_decl;
	term_tab_decl = php_mode_term_tab_decl;
	statement_sep = php_mode_statement_sep;
	out_err = php_mode_out_err;
	string_delim = php_mode_string_delim;
}

void usage()
{
	printf("%s: %s [-s] [-d dbg] [-o outfile] file\n", prognam, prognam);
	printf("\n");
	printf("-s\tdisallow duplicate keys\n");
	printf("-d dbg\tdebug flags\n");
	printf("-o outfile\toutput file\n");
	printf("\n");
	printf("Debug flags are a bitwise combination of:\n");
	printf("1\tdump trie\n");
	exit(0);
}

char *cstr(char *s)
{
	char *r;
	char *rp;
	
	if (!s)
		s = "";
	r = calloc(strlen(s)+1, sizeof(*r));
	memchk(r);
	rp = r;
	while (*s)
	{
		switch(*s)
		{
			case '\\':
				s++;
				switch(*s)
				{
					case 'n':
						*rp = '\n';
						break;
					case 'r':
						*rp = '\r';
						break;
					case 't':
						*rp = '\t';
						break;
					case '\\':
						*rp = '\\';
						break;
					default:
						fprintf(stderr, "%s: Invalid escape at line %d\n", prognam, lineno);
						exit(1);
				}
				break;
			default:
				*rp = *s;
		}
		rp++;
		s++;
	}
	*rp = '\0';
	return(r);
}

int main(int argc, char *argv[])
{
	char *buf;
	char *s1;
	char *s2;
	struct trieTab *trie_tab;
	struct termTab *term_tab;
	int ch;
	int inliteral = 0;

	fin = stdin;
	fout = stdout;

	buf = calloc(bufsiz, sizeof(*buf));

	while ((ch = getopt(argc, argv, "sd:o:")) != -1)
	{
		switch(ch)
		{
			case 's':
				no_duplicates = 1;
				break;
			case 'd':
				debug_mode = atoi(optarg);
				break;
			case 'o':
				fout = fopen(optarg, "wt");
				if (!fout)
				{
					fprintf(stderr, "%s: %s: %s\n", prognam, optarg, strerror(errno));
					exit(1);
				}
				break;
			case '?':
			default:
				usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (argc > 0)
	{
		fin = fopen(argv[0], "rt");
		if (!fin)
		{
			fprintf(stderr, "%s: %s: %s\n", prognam, argv[0], strerror(errno));
			exit(1);
		}
	}
	/* definition section */
	linein(buf,fin);
	while (!feof(fin))
	{
		if (!strncmp(buf,SEP,SEPLEN))
			break;
		if (!strncmp(buf,LITBEGIN,LITLEN))
			inliteral = 1;
		else if (!strncmp(buf,LITEND,LITLEN))
			inliteral = 0;
		else if (inliteral)
			fprintf(fout,"%s", buf);
		else if (strlen(buf) < 3 && (buf[0] == '\r' || buf[0] == '\n'))
		{
			/* ignore blank lines */
		}
		else if (buf[0] == '%')
		{
			/* handle directives */
			s1 = strtok(buf+1, " \t\n");
			s2 = strtok(NULL, "\n");
			if (!strcmp(s1,"bufsiz"))
			{
				bufsiz = atoi(s2);
				if (bufsiz < 12)
				{
					fprintf(stderr, "%s: bufsiz dangerously low at line %d\n", prognam, lineno);
					exit(1);
				}
				buf = realloc(buf, bufsiz * sizeof(*buf));
				memchk(buf);
			}
			else if (!strcmp(s1,"comment_open"))
			{
				comment_open = cstr(s2);
			}
			else if (!strcmp(s1,"comment_close"))
			{
				comment_close = cstr(s2);
			}
			else if (!strcmp(s1,"trie_tab_decl"))
			{
				trie_tab_decl = cstr(s2);
			}
			else if (!strcmp(s1,"term_tab_decl"))
			{
				term_tab_decl = cstr(s2);
			}
			else if (!strcmp(s1,"string_delim"))
			{
				string_delim = cstr(s2);
			}
			else if (!strcmp(s1,"out_tab_decl"))
			{
				out_tab_decl = cstr(s2);
			}
			else if (!strcmp(s1,"statement_sep"))
			{
				statement_sep = cstr(s2);
			}
			else if (!strcmp(s1,"column_open"))
			{
				column_open = cstr(s2);
			}
			else if (!strcmp(s1,"column_close"))
			{
				column_close = cstr(s2);
			}
			else if (!strcmp(s1,"column_sep"))
			{
				column_sep = cstr(s2);
			}
			else if (!strcmp(s1,"row_open"))
			{
				row_open = cstr(s2);
			}
			else if (!strcmp(s1,"row_close"))
			{
				row_close = cstr(s2);
			}
			else if (!strcmp(s1,"row_sep"))
			{
				row_sep = cstr(s2);
			}
			else if (!strcmp(s1,"out_err"))
			{
				out_err = cstr(s2);
			}
			else if (!strcmp(s1,"mode"))
			{
				if (!strcmp("c", s2))
					c_mode();
				else if (!strcmp("perl", s2))
					perl_mode();
				else if (!strcmp("php", s2))
					php_mode();
				else
				{
					fprintf(stderr, "%s: Invalid mode at line %d\n", prognam, lineno);
					exit(1);
				}
			}
			else
			{
				fprintf(stderr, "%s: Invalid directive at line %d\n", prognam, lineno);
				exit(1);
			}
		}
		else
		{
			fprintf(stderr, "%s: Syntax error at line %d\n", prognam, lineno);
			exit(1);
		}
		linein(buf,fin);
	}
	/* rules section */
	linein(buf,fin);
	while (!feof(fin))
	{
		if (strlen(buf) < 3 && (buf[0] == '\r' || buf[0] == '\n'))
			goto nextline;
		if (!strncmp(buf,SEP,SEPLEN))
			break;
		s1 = strtok(buf, " \t");
		s2 = strtok(NULL, "\r\n");
		if (!strlen(s1) || s1[0] == '\n' || s1[0] == '\r')
		{
			fprintf(stderr, "%s: Zero length pattern at line %d\n", prognam, lineno);
			exit(1);
		}
		if (strspn(s1, "0123456789") != strlen(s1))
		{
			fprintf(stderr, "%s: Pattern contained non digit at line %d\n", prognam, lineno);
			exit(1);
		}
		top = insertTrie(top, 0, s1, s2);
nextline:
		linein(buf,fin);
	}
	propagateLeaves(top);
	if (debug_mode & 0x1)
	{
		fprintf(fout,"\n");
		dump_trie(top);
		fprintf(fout,"\n");
	}
	fprintf(fout, comment_open);
	fprintf(fout,"Output table");
	fprintf(fout, comment_close);
	fprintf(fout,"\n");
	dump_strtab(strtab);
	fprintf(fout, comment_open);
	fprintf(fout,"Trie table");
	fprintf(fout, comment_close);
	fprintf(fout,"\n");
	trie_tab = calloc(nicnt, sizeof(*trie_tab));
	term_tab = calloc(ntcnt, sizeof(*term_tab));
	trie_map(top, trie_tab,term_tab);
	dump_trie_tab(trie_tab, nicnt);
	dump_term_tab(term_tab, ntcnt);
	fprintf(fout,"\n");
	/* user subroutine section */
	linein(buf,fin);
	while (!feof(fin))
	{
		fprintf(fout,"%s", buf);
		linein(buf,fin);
	}
	return 0;
}
